/* KinokoIndexedDataSection.hh */
/* Created by Enomoto Sanshiro on 19 November 2000. */
/* Last updated by Enomoto Sanshiro on 19 November 2000. */


#ifndef __KinokoIndexedDataSection_hh__
#define __KinokoIndexedDataSection_hh__

#include <iostream>
#include <string>
#include "ParaTokenizer.hh"
#include "KinokoDataSection.hh"


class TKinokoIndexedDataSectionFormatter;
class TKinokoIndexedDataSectionScanner;


class TKinokoIndexedDataSection: public TKinokoDataSection {
  public:
    TKinokoIndexedDataSection(TKinokoDataSource* DataSource, const std::string& SectionName);
    TKinokoIndexedDataSection(TKinokoDataSource* DataSource, const std::string& SectionName, int SectionId);
    virtual ~TKinokoIndexedDataSection();
    virtual int SectionType(void) const;
    virtual void ReadFrom(TParaTokenizer& Tokenizer) throw(TKinokoException);
    virtual void WriteTo(std::ostream& os, const std::string& Indent);
    inline TKinokoIndexedDataSectionFormatter* Formatter(void);
    inline TKinokoIndexedDataSectionScanner* Scanner(void);
  public:
    void SetWidth(unsigned AddressWidth, unsigned DataWidth);
    void SetElementDepth(unsigned ElementDepth);
    unsigned AddressWidth(void) const;
    unsigned DataWidth(void) const;
    unsigned ElementDepth(void) const;
  protected:
    unsigned _AddressWidth, _DataWidth, _ElementDepth;
    TKinokoIndexedDataSectionFormatter* _Formatter;
    TKinokoIndexedDataSectionScanner* _Scanner;
};


class TKinokoIndexedDataSectionFormatter: public TKinokoDataSectionFormatter {
  public:
    TKinokoIndexedDataSectionFormatter(TKinokoIndexedDataSection* DataSection);
    virtual ~TKinokoIndexedDataSectionFormatter();
    inline int DataSizeFor(int NumberOfElements) const;
    inline int WriteTo(void* Buffer, int Index, int Address, int Data) const;
    inline int WriteArrayTo(void* Buffer, int StartIndex, int Address, int DataArray[], int DataArrayLength) const;
  protected:
    unsigned _AddressWidth, _DataWidth;
    bool _UsesTwoWords;
  public:
    enum {
	Offset_Base = TKinokoDataSectionFormatter::Offset_DataArea
    };
};


class TKinokoIndexedDataSectionScanner: public TKinokoDataSectionScanner {
  public:
    TKinokoIndexedDataSectionScanner(TKinokoIndexedDataSection* DataSection);
    virtual ~TKinokoIndexedDataSectionScanner();
    inline int NumberOfElements(void* Buffer) const;
    inline void ReadFrom(void* Buffer, int Index, int& Address, int& Data) const;
  protected:
    unsigned _AddressWidth, _DataWidth, _ElementWidth;
  public:
    enum {
	Offset_DataAreaSize = TKinokoDataSectionFormatter::Offset_DataAreaSize,
	Offset_Base = TKinokoDataSectionFormatter::Offset_DataArea
    };
};



inline TKinokoIndexedDataSectionFormatter* TKinokoIndexedDataSection::Formatter(void)
{
    if (_Formatter == 0) {
	_Formatter = new TKinokoIndexedDataSectionFormatter(this);
    }

    return _Formatter;
}

inline TKinokoIndexedDataSectionScanner* TKinokoIndexedDataSection::Scanner(void)
{
    if (_Scanner == 0) {
	_Scanner = new TKinokoIndexedDataSectionScanner(this);
    }

    return _Scanner;
}



inline int TKinokoIndexedDataSectionFormatter::DataSizeFor(int NumberOfElements) const
{
    if (_UsesTwoWords) {
	return sizeof(U32bit) * (2 * NumberOfElements);
    }
    else {
	return sizeof(U32bit) * NumberOfElements;
    }
}

inline int TKinokoIndexedDataSectionFormatter::WriteTo(void* Buffer, int Index, int Address, int Data) const
{
    if (_UsesTwoWords) {
	((U32bit*) Buffer)[Offset_Base + 2 * Index] = Address;
	((U32bit*) Buffer)[Offset_Base + 2 * Index + 1] = Data;
	return 2 * sizeof(U32bit);
    }
    else {
	((U32bit*) Buffer)[Offset_Base + Index] = Address << _DataWidth | Data;
	return sizeof(U32bit);
    }
}

inline int TKinokoIndexedDataSectionFormatter::WriteArrayTo(void* Buffer, int StartIndex, int Address, int DataArray[], int DataArrayLength) const
{
    if (_UsesTwoWords) {
	for (int i = 0; i < DataArrayLength; i++) {
	    ((U32bit*) Buffer)[Offset_Base + 2 * (StartIndex+i)] = Address;
	    ((U32bit*) Buffer)[Offset_Base + 2 * (StartIndex+i) + 1] = DataArray[i];
	}
	return 2 * sizeof(U32bit) * DataArrayLength;
    }
    else {
	for (int i = 0; i < DataArrayLength; i++) {
	    ((U32bit*) Buffer)[Offset_Base + (StartIndex + i)] = Address << _DataWidth | DataArray[i];
	}
	return sizeof(U32bit) * DataArrayLength;
    }
}



inline int TKinokoIndexedDataSectionScanner::NumberOfElements(void* Buffer) const
{
    return (8 * ((U32bit*) Buffer)[Offset_DataAreaSize]) / _ElementWidth;
}

inline void TKinokoIndexedDataSectionScanner::ReadFrom(void* Buffer, int Index, int& Address, int& Data) const
{
    // I know the packing order is somewhat awkward;
    // this is for backward compatibility.

    U32bit* DataBlock = (U32bit*) Buffer + Offset_Base;
    unsigned StartBit, WordIndex, BitIndex;
    int CarryOverWidth;

    StartBit = Index * _ElementWidth;
    WordIndex = StartBit / 32;
    BitIndex = StartBit % 32;
    CarryOverWidth = BitIndex + _AddressWidth - 32;
    if (CarryOverWidth <= 0) {
	Address = DataBlock[WordIndex] >> (-CarryOverWidth);
    }
    else {
	Address = DataBlock[WordIndex] << CarryOverWidth;
	Address |= DataBlock[WordIndex + 1] >> (32 - CarryOverWidth);
    }

    if (_AddressWidth < 32) {
	Address &= (0x00000001u << _AddressWidth) - 1;
    }
    else {
	Address &= 0xffffffff;
    }

    StartBit = Index * _ElementWidth + _AddressWidth;
    WordIndex = StartBit / 32;
    BitIndex = StartBit % 32;
    CarryOverWidth = BitIndex + _DataWidth - 32;
    if (CarryOverWidth <= 0) {
	Data = DataBlock[WordIndex] >> (-CarryOverWidth);
    }
    else {
	Data = DataBlock[WordIndex] << CarryOverWidth;
	Data |= DataBlock[WordIndex + 1] >> (32 - CarryOverWidth);
    }

    if (_DataWidth < 32) {
	Data &= (0x00000001u << _DataWidth) - 1;
    }
    else {
	Data &= 0xffffffff;
    }
}


#endif
