/* KinokoControl.cc */
/* Created by Enomoto Sanshiro on 1 October 2001. */
/* Last updated by Enomoto Sanshiro on 27 March 2002. */


#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>
#include <string>
#include <vector>
#include <map>
#include <set>
#include "MushXml.hh"
#include "MushFileSystem.hh"
#include "MushMisc.hh"
#include "KinokoShell.hh"
#include "KinokoControlWidget.hh"
#include "KinokoControlPanel.hh"
#include "KinokoControlScript.hh"
#include "KinokoControl.hh"

using namespace std;
using namespace mush;


TKinokoControl::TKinokoControl(TKinokoShellConnector* ShellConnector)
: TKinokoShell(ShellConnector)
{
    _Script = 0;
    _ShellScript = 0;

    _CurrentState = "initial";
}

TKinokoControl::~TKinokoControl()
{
    std::map<string, TKinokoControlWidget*>::iterator WidgetIterator;
    for (
	WidgetIterator = _WidgetTable.begin();
	WidgetIterator != _WidgetTable.end();
	WidgetIterator++
    ){
	delete (*WidgetIterator).second;
    }

    delete _Script;
    delete _ShellScript;
}

void TKinokoControl::AddWidget(TKinokoControlWidget* Widget, const string& EnabledStateList)
{
    _WidgetTable[Widget->Name()] = Widget;

    if (! EnabledStateList.empty()) {
	string State;
	set<string> StateSet;
	istringstream StateListStream(EnabledStateList);
	while (StateListStream >> State) {
	    StateSet.insert(State);
	}

	_EnabledStateList[Widget] = StateSet;
    }
}

void TKinokoControl::AddInputWidget(TKinokoControlWidget* Widget)
{
    _InputWidgetList.push_back(Widget);
    _InputValueCache.push_back("");
}

void TKinokoControl::LoadControlScript(istream& ScriptText)
{
    if (_Script == 0) {
	if (_ShellScript == 0) {
	    _ShellScript = new TKinokoShellScript(
		CreatePopupWindow(), CreateFileSelectDialog()
	    );
	}
	_Script = new TKinokoControlScript(this);
	_Script->Merge(_ShellScript);
    }

    try {
        _Script->Parse(ScriptText);
	_Script->Execute();
    }
    catch (TScriptException &e) {
	HandleException(
	    TKinokoShellException("ScriptException: " + e.Message())
	);
    }
}

void TKinokoControl::OnClickWidget(TKinokoControlWidget* Widget)
{
    string Action = Widget->ActionOnClick();
    if (Action.empty()) {
	Invoke(Widget->Name());
    }
    else {
	ExecuteAction(Action);
    }
}

int TKinokoControl::Invoke(const string& MethodName)
{
    if (_Script != 0) {
	try {
	    _Script->ExecuteInvocationTask(MethodName);
	}
	catch (TScriptException &e) {
	    HandleException(
		TKinokoShellException("ScriptException: " + e.Message())
	    );
	}
    }

    ostringstream MessageStream;
    MessageStream << ".invoke " << MethodName << " ";
    for (unsigned i = 0; i < _InputWidgetList.size(); i++) {
	string Name = _InputWidgetList[i]->Name();
	string Value = ReplaceEscape(_InputWidgetList[i]->Value());

	if (Value != _InputValueCache[i]) {
	    _InputValueCache[i] = Value;
	    MessageStream << Name << "=" << Value << "&";
	}
    }
    MessageStream << ";" << endl;

    string Message = MessageStream.str();

    _ShellConnector->SendMessage((char*) Message.c_str(), Message.size());

    return 1;
}

int TKinokoControl::ExecuteAction(const string& ActionName)
{
    try {
	_Script->Execute(ActionName);
    }
    catch (TScriptException &e) {
	HandleException(
	    TKinokoShellException("ScriptException: " + e.Message())
        );
    }

    return 1;
}

int TKinokoControl::OnStartup(void)
{
    if (_Script == 0) {
	return 0;
    }

    try {
	vector<TParaValue*> ArgumentList;
	_Script->ExecuteEventTask("startup", ArgumentList);
    }
    catch (TScriptException &e) {
	HandleException(
	    TKinokoShellException("ScriptException: " + e.Message())
        );
    }

    return 1;
}

int TKinokoControl::OnShutdown(void)
{
    if (_Script == 0) {
	return 0;
    }

    try {
	vector<TParaValue*> ArgumentList;
	_Script->ExecuteEventTask("shutdown", ArgumentList);
    }
    catch (TScriptException &e) {
	HandleException(
	    TKinokoShellException("ScriptException: " + e.Message())
	);
    }

    return 1;
}

int TKinokoControl::OnEverySecond(void)
{
    if (_Script == 0) {
	return 0;
    }

    try {
	_Script->ExecuteScheduledTask(TMushDateTime::SecSinceEpoch());
    }
    catch (TScriptException &e) {
	HandleException(
	    TKinokoShellException("ScriptException: " + e.Message())
	);
    }

    return 1;
}

int TKinokoControl::ProcessSystemCommand(const string& Command, std::istream& InputStream)
{
    int Result = 0;
    if (Command == ".openControlPanel") {
	Result = ProcessOpenControlPanelCommand(InputStream);
    }
    else if (Command == ".clearWidgetValues") {
	Result = ProcessClearWidgetValuesCommand(InputStream);
    }
    else if (Command == ".saveWidgetValues") {
	Result = ProcessSaveWidgetValuesCommand(InputStream);
    }
    else if (Command == ".loadWidgetValues") {
	Result = ProcessLoadWidgetValuesCommand(InputStream);
    }
    else if (Command == ".setWidgetValue") {
	Result = ProcessSetWidgetValueCommand(InputStream);
    }
    else if (Command == ".loadControlScript") {
	Result = ProcessLoadControlScriptCommand(InputStream);
    }
    else if (Command == ".executeAction") {
	Result = ProcessExecuteActionCommand(InputStream);
    }
    else if (Command == ".changeState") {
	Result = ProcessChangeStateCommand(InputStream);
    }
    else if (Command == ".openPopup") {
	Result = ProcessOpenPopupCommand(InputStream, false);
    }
    else if (Command == ".openQueryPopup") {
	Result = ProcessOpenPopupCommand(InputStream, true);
    }
    else {
	Result = TKinokoShell::ProcessSystemCommand(Command, InputStream);
    }

    return Result;
}

int TKinokoControl::ProcessOpenControlPanelCommand(istream& InputStream)
{
    string ControlPanelScriptFileName;
    if (InputStream >> ControlPanelScriptFileName) {
	Construct(ControlPanelScriptFileName);
    }

    return 1;
}

int TKinokoControl::ProcessClearWidgetValuesCommand(istream& InputStream)
{
    ClearWidgetValues();

    return 1;
}

int TKinokoControl::ProcessSaveWidgetValuesCommand(istream& InputStream)
{
    string FileName;
    if (InputStream >> FileName) {
	SaveWidgetValues(FileName);
    }

    return 1;
}

int TKinokoControl::ProcessLoadWidgetValuesCommand(istream& InputStream)
{
    string FileName;
    if (InputStream >> FileName) {
	LoadWidgetValues(FileName);
    }

    return 1;
}

int TKinokoControl::ProcessSetWidgetValueCommand(istream& InputStream)
{
    string WidgetName;
    if (InputStream >> WidgetName >> ws) {
	char Value[256];
	InputStream.getline(Value, sizeof(Value), ';');
	SetWidgetValue(WidgetName, Value);
    }

    return 1;
}

int TKinokoControl::ProcessExecuteActionCommand(istream& InputStream)
{
    int Result = 0;

    string ActionName;
    if (InputStream >> ActionName) {
	Result = ExecuteAction(ActionName);
    }

    return Result;
}

int TKinokoControl::ProcessLoadControlScriptCommand(istream& InputStream)
{
    int Result = 0;

    string ControlScriptFileName;
    if (InputStream >> ControlScriptFileName) {
	ifstream ControlScriptText(ControlScriptFileName.c_str());
	if (ControlScriptText) {
	    LoadControlScript(ControlScriptText);
	    Result = 1;
	}
    }

    return Result;
}

int TKinokoControl::ProcessChangeStateCommand(istream& InputStream)
{
    string StateName;
    if (InputStream >> StateName) {
	ChangeState(StateName);
    }

    return 1;
}

int TKinokoControl::ProcessOpenPopupCommand(istream& InputStream, bool IsQuery)
{
    TKinokoShellPopupWindow* PopupWindow = CreatePopupWindow();

    string Type;
    if (getline(InputStream >> ws, Type, ',')) {
	PopupWindow->SetType(Type);
    }

    string ActionList;
    if (getline(InputStream >> ws, ActionList, ',')) {
	istringstream ActionListStream(ActionList);
	string Action;
	while (ActionListStream >> Action) {
	    PopupWindow->AddAction(Action);
	}
    }

    string Message;
    getline(InputStream >> ws, Message, ';');
    PopupWindow->SetMessage(Message.c_str());

    bool IsModal = IsQuery;
    string Selection = PopupWindow->Open(IsModal);

    if (IsQuery) {
	string Reply = Selection + ";\n";
	_ShellConnector->SendMessage((char*) Reply.c_str(), Reply.size());
    }

    delete PopupWindow;

    return 1;
}

int TKinokoControl::ChangeState(const string& StateName)
{
    map<TKinokoControlWidget*, set<std::string> >::iterator i;
    for (i = _EnabledStateList.begin(); i != _EnabledStateList.end(); i++) {
	TKinokoControlWidget* Widget = (*i).first;

	if ((*i).second.count(StateName) > 0) {
	    Widget->Enable();
	}
	else {
	    Widget->Disable();
	}
    }

    if (_Script != 0) {
	try {
	    _Script->ExecuteTransitionTask(_CurrentState, StateName);
	}
	catch (TScriptException &e) {
	    HandleException(
		TKinokoShellException("ScriptException: " + e.Message())
	    );
	}
    }

    _CurrentState = StateName;

    return 1;
}

int TKinokoControl::SetWidgetValue(const string& WidgetName, const string& Value)
{
    if (_WidgetTable.count(WidgetName) == 0) {
	return 0;
    }
    _WidgetTable[WidgetName]->SetValue(Value);

    return 1;
}

int TKinokoControl::ClearWidgetValues(void)
{
    for (unsigned i = 0; i < _InputWidgetList.size(); i++) {
	_InputWidgetList[i]->SetValue("");
    }

    return 1;
}

int TKinokoControl::SaveWidgetValues(const string& FileName)
{
    ofstream File(FileName.c_str());
    if (! File) {
	return 0;
    }

    File << "<?xml version=\"1.0\"?>" << endl;
    File << endl;
    File << "<kcml-widget-values>" << endl;

    for (unsigned i = 0; i < _InputWidgetList.size(); i++) {
	File << "    <widget name=\"" << _InputWidgetList[i]->Name() << "\">";
	File << _InputWidgetList[i]->Value();
	File << "</widget>";
	File << endl;
    }
    File << "</kcml-widget-values>" << endl;

    return 1;
}

int TKinokoControl::LoadWidgetValues(const string& FileName)
{
    class TDocumentHandler: public sax::DocumentHandler {
      public:
	TDocumentHandler(TKinokoControl* Control): _Control(Control) {}
	virtual ~TDocumentHandler() {}
      public:
	virtual void startElement(
	    const std::string& name, 
	    const sax::AttributeList& attribute_list
	){
	    if (name == "widget") {
		_Name = attribute_list.getValue("name");
		_Value = "";
	    }
	}
	virtual void endElement(const std::string& name) 
        {
	    if (name == "widget") {
		_Control->SetWidgetValue(_Name, _Value);
	    }
	}
	virtual void characters(const std::string& text) 
        {
	    _Value += text;
	}
      protected:
	TKinokoControl* _Control;
	std::string _Name, _Value;
    };

    if (! TMushFileAttribute(FileName).IsReadable()) {
	return 0;
    }

    try {
	TDocumentHandler Handler(this);
	mush::sax::Parser Parser;
	Parser.setDocumentHandler(&Handler);
	Parser.parse(FileName);
    }
    catch (TSystemCallException &e) {
	cerr << "ERROR: " << e << endl;
	return 0;
    }

    return 1;
}

const std::string& TKinokoControl::CurrentState(void)
{
    return _CurrentState;
}

TKinokoControlWidget* TKinokoControl::LookupWidget(const std::string& WidgetName)
{
    return _WidgetTable[WidgetName];
}

string TKinokoControl::ReplaceEscape(const string& Value)
{
    string Result;
    for (unsigned i = 0; i < Value.size(); i++) {
	switch (Value[i]) {
	  case '%':
	    Result += "%p";
	    break;
	  case '&':
	    Result += "%a";
	    break;
          case '=':
	    Result += "%e";
	    break;
	  case ';':
	    Result += "%s";
	    break;
	  default:
	    Result += Value[i];
	}
    }

    return Result;
}
